""" Tests metadata base class. """

import os
import glob
import unittest
import datetime

from src.medali.core import MetaData


class MetadataConfigReadTest(unittest.TestCase):
    """ Tests reading of all available config files. """

    def test_configs_read(self):
        """ Test reading all available config files and filling null values. """
        root_dirpath = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "src", "medali"))
        lib_path = os.path.join(root_dirpath, "lib")

        product_names = os.listdir(lib_path)
        cfg_filepaths = []
        for product_name in product_names:
            product_dirpath = os.path.join(lib_path, product_name)
            cfg_filepaths.extend(glob.glob(os.path.join(product_dirpath, "*.ini")))

        for cfg_filepath in cfg_filepaths:
            metadata = MetaData.from_cfg_file({}, cfg_filepath)
            self.assertEqual(metadata._meta.keys(), metadata._ref_meta['Metadata'].keys())


class MetadataTest(unittest.TestCase):
    """
    Test reading, checking and converting metadata with respect to the reference metadata
    and other metadata classes.

    """

    def setUp(self):
        """ Reads template config metadata file and creates a `MetaData` instance. """
        test_data_dirpath = os.path.join(os.path.dirname(__file__), "test_data")
        cfg_filepath = os.path.join(test_data_dirpath, "cfg_template.ini")
        self.cfg_filepath = cfg_filepath
        self.metadata = MetaData.from_cfg_file({}, cfg_filepath)
        self.metadata['datetime_type'] = datetime.datetime(2020, 12, 12, 12, 20, 10)
        self.metadata['boolean_type'] = False
        self.metadata['number_type'] = 1.2
        self.metadata['integer_type'] = 1
        self.metadata['string_general'] = 'abc'
        self.metadata['string_list'] = 'V3'
        self.metadata['string_pattern'] = 'Pattern01'

    def test_set_metadata_attributes(self):
        """ Tests metadata item setting. """
        metadata_should = {'datetime_type': '2020-12-12 12:20:10',
                           'string_pattern': 'Pattern01',
                           'string_general': 'abc',
                           'integer_type': '1',
                           'number_type': '1.2',
                           'string_list': 'V3',
                           'boolean_type': 'False'}

        self.assertDictEqual(metadata_should, self.metadata._meta)

    def test_get_metadata_attributes(self):
        """ Tests value decoding when accessing metadata attributes. """

        assert self.metadata['datetime_type'] == datetime.datetime(2020, 12, 12, 12, 20, 10)
        assert not self.metadata['boolean_type']
        assert self.metadata['number_type'] == 1.2
        assert self.metadata['integer_type'] == 1
        assert self.metadata['string_general'] == 'abc'
        assert self.metadata['string_list'] == 'V3'
        assert self.metadata['string_pattern'] == 'Pattern01'

    def test_expected_values(self):
        """ Tests checking of external metadata. """

        metadata = MetaData.from_cfg_file({}, self.cfg_filepath)
        metadata['datetime_type'] = '2020-12-12 12:20:10'
        metadata['string_pattern'] = 'null'
        metadata['integer_type'] = '1'
        try:
            metadata['number_type'] = 'haha'
            assert False
        except ValueError:
            assert True
        try:
            metadata['string_list'] = 'V5'
            assert False
        except ValueError:
            assert True
        metadata['boolean_type'] = 'False'

    def test_and(self):
        """ Tests AND operation between two `MetaData` instances. """

        ref_dtypes = {'datetime_type': 'datetime',
                      'boolean_type': 'boolean',
                      'number_type': 'number',
                      'integer_type': 'integer'}
        ref_metadata = dict()
        ref_metadata['Metadata'] = ref_dtypes
        sec_metadata = MetaData({}, ref_metadata)

        metadata_should = {'datetime_type': '2020-12-12 12:20:10',
                           'integer_type': '1',
                           'number_type': '1.2',
                           'boolean_type': 'False'}

        common_metadata = self.metadata & sec_metadata

        self.assertDictEqual(metadata_should, common_metadata._meta)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
    Setup file for medali.
    Use setup.cfg to configure your project.

    This file was generated with PyScaffold 3.3.
    PyScaffold helps you to put up the scaffold of your new Python project.
    Learn more under: https://pyscaffold.org/
"""
import sys
from pkg_resources import VersionConflict, require
from setuptools import setup

try:
    require("setuptools>=38.3")
except VersionConflict:
    print("Error: version of setuptools is too old (<38.3)!")
    sys.exit(1)


if __name__ == "__main__":
    setup(use_pyscaffold=True)



""" Parsing and modification of metadata. """

import re
import os
import numbers
import datetime
from dateutil.parser import parse
from pprint import pformat
from configparser import ConfigParser


class MetaData:
    """ Metadata base class to store metadata, set and get metadata. """
    def __init__(self, metadata, ref_metadata=None):
        """
        Creates a `MetaData` instance from a given metadata dictionary
        and a dictionary storing information about expected metadata
        attributes, metadata value data types, and expected metadata values.

        Parameters
        ----------
        metadata : dict
            Dictionary containing metadata attributes and decoded values.
        ref_metadata : dict
            Dictionary containing expected metadata attributes plus data types
            under the key "Metadata", and expected metadata values under the key
            "Expected_value".

        """
        self._meta = {}
        self._ref_meta = {'Metadata': dict(), 'Expected_value': dict()} if ref_metadata is None else ref_metadata
        self._set_input_metadata(metadata)

    @classmethod
    def from_cfg_file(cls, metadata, cfg_filepath):
        """
        Creates a `MetaData` instance from a given metadata dictionary and
        a config file.

        Parameters
        ----------
        metadata : dict
            Dictionary containing metadata attributes and decoded values.
        cfg_filepath : str
            Path to metadata config file.

        Returns
        -------
        MetaData

        """
        if not os.path.exists(cfg_filepath):
            err_msg = "'{}' does not exist.".format(cfg_filepath)
            raise IOError(err_msg)
        ref_metadata = read_config(cfg_filepath)

        return cls(metadata, ref_metadata)

    @classmethod
    def from_product_version(cls, metadata, prod_name, version_id):
        """
        Creates a `MetaData` instance from a given metadata dictionary,
        a product name and a version ID.

        Parameters
        ----------
        metadata : dict
            Dictionary containing metadata attributes and decoded values.
        prod_name : str
            Name of the product, e.g. "SIG0".
        version_id : str
            Metadata version.

        Returns
        -------
        MetaData

        """
        root_dirpath = os.path.dirname(os.path.abspath(__file__))
        cfg_filepath = os.path.join(root_dirpath, "lib", prod_name.lower(), version_id + '.ini')

        return cls.from_cfg_file(metadata, cfg_filepath)

    def to_pretty_frmt(self):
        """ str : Returns metadata dictionary in a formatted string. """
        return pformat(self._meta, indent=4)

    def to_tags(self):
        """ dict : Returns metadata as a dictionary containing encoded values. """
        return self._meta

    def _set_input_metadata(self, metadata):
        """
        Sets metadata attributes and values according to the given metadata dictionary.
        If a required attribute is not given it is set to 'null'.

        Parameters
        ----------
        metadata : dict
            Dictionary containing metadata attributes and decoded values.

        """
        for key, value in metadata.items():
            self._set_metadata(key, value)

        for key, value in self._ref_meta['Metadata'].items():
            if key not in self._meta:
                self._meta[key] = 'null'

    def _set_metadata(self, attr, value):
        """
        Encodes the given metadata value according to the given metadata attribute and
        stores it/overwrites it.

        Parameters
        ----------
        attr : str
            Metadata attribute.
        value : any
            Metadata value.

        """
        # convert back and forth to execute expected value test with decoded values
        enc_value = self._encode(attr, value)
        dec_value = self._decode(attr, enc_value)
        if not self._is_expected(attr, dec_value):
            expected_values = self._ref_meta['Expected_value'].get(attr)
            err_msg = "Metadata value '{}' is not in compliance with '{}'".format(value, expected_values)
            raise ValueError(err_msg)

        self._meta[attr] = self._encode(attr, value)

    def _get_metadata(self, attr):
        """
        Decodes and returns metadata value according to the given attribute.

        Parameters
        ----------
        attr : str
            Metadata attribute.

        Returns
        -------
        value : any
            Decoded metadata value.

        """
        if attr not in self._meta.keys():
            err_msg = "Metadata attribute '{}' can not be found.".format(attr)
            raise KeyError(err_msg)

        return self._decode(attr, self._meta[attr])

    def _is_expected(self, attr, value):
        """
        Checks if a given metadata value is in compliance with the
        corresponding expected values in the reference metadata.

        Parameters
        ----------
        attr : str
            Metadata attribute.
        value : any
            Decoded metadata value.

        Returns
        ----------
        is_expected : bool
            True if the given metadata value is expected, else false.

        """
        exp_values = self._ref_meta['Expected_value'].get(attr)
        is_expected = True
        if exp_values and (value not in [None, 'null']):
            if isinstance(exp_values, list):
                if value not in exp_values:
                    is_expected = False
            elif exp_values.startswith('pattern'):
                exp_value = exp_values.replace(', ', ',')
                pattern = exp_value.split(',')[1]
                if not re.search(pattern, value):
                    is_expected = False

        return is_expected

    def _decode(self, attr, value):
        """
        Decodes value according to the given attribute.

        Parameters
        ----------
        attr : str
            Metadata attribute.
        value : str
            Metadata value.

        Returns
        -------
        dec_value : any
            Decoded metadata value.

        """
        if self._ref_meta['Metadata']:
            if attr in self._ref_meta['Metadata'].keys():
                if value == 'none':
                    dec_value = None
                elif value == 'null':
                    dec_value = 'null'
                elif self._ref_meta['Metadata'][attr] == 'boolean':
                    dec_value = value == 'True'
                elif self._ref_meta['Metadata'][attr] == 'integer':
                    dec_value = int(value)
                elif self._ref_meta['Metadata'][attr] == 'number':
                    dec_value = float(value)
                elif self._ref_meta['Metadata'][attr] == 'datetime':
                    dec_value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
                else:
                    dec_value = value
            else:
                err_msg = "Attribute '{}' is not given in the reference metadata."
                raise KeyError(err_msg)
        else:
            dec_value = value

        return dec_value

    def _encode(self, attr, value):
        """
        Encodes value according to the given attribute.

        Parameters
        ----------
        attr : str
            Metadata attribute.
        value : any
            Metadata value.

        Returns
        -------
        enc_value : str
            Encoded metadata value.

        """
        err_msg_frmt = "Metadata value for attribute '{}' has to be '{}'."
        if self._ref_meta['Metadata']:
            if attr in self._ref_meta['Metadata'].keys():
                dtype = self._ref_meta['Metadata'][attr]
                if value is None:
                    enc_value = 'none'
                elif isinstance(value, str):  # nothing to encode, but check if the value is convertable
                    try:
                        self._decode(attr, value)
                    except Exception:
                        raise ValueError(err_msg_frmt.format(attr, dtype))
                    enc_value = value
                elif dtype == 'boolean':
                    if isinstance(value, bool):
                        enc_value = str(value)
                    else:
                        raise ValueError(err_msg_frmt.format(attr, dtype))
                elif dtype == 'integer':
                    if isinstance(value, int):
                        enc_value = str(value)
                    else:
                        raise ValueError(err_msg_frmt.format(attr, dtype))
                elif dtype == 'number':
                    if isinstance(value, numbers.Number):
                        enc_value = str(value)
                    else:
                        raise ValueError(err_msg_frmt.format(attr, dtype))
                elif dtype == 'datetime':
                    if isinstance(value, datetime.datetime):
                        enc_value = value.strftime("%Y-%m-%d %H:%M:%S")
                    else:
                        raise ValueError(err_msg_frmt.format(attr, dtype))
                else:
                    enc_value = str(value)
            else:
                err_msg = "Attribute '{}' is not given in the reference metadata.".format(attr)
                raise KeyError(err_msg)
        else:
            enc_value = value

        return enc_value

    def __and__(self, other):
        """ Finds common metadata attributes among the two metadata classes. """
        common_keys = self._meta.keys() & other._meta.keys()
        common_metadata = dict()
        common_ref_metadata = dict()
        common_ref_metadata['Metadata'] = dict()
        common_ref_metadata['Expected_value'] = dict()
        for common_key in common_keys:
            common_metadata[common_key] = self._meta[common_key]
            ref_meta = self._ref_meta['Metadata'].get(common_key)
            if ref_meta is not None:
                common_ref_metadata['Metadata'][common_key] = ref_meta
            ref_exp_value = self._ref_meta['Expected_value'].get(common_key)
            if ref_exp_value is not None:
                common_ref_metadata['Expected_value'][common_key] = ref_exp_value

        return MetaData(common_metadata, common_ref_metadata)

    def __str__(self):
        """ str : String representation of metadata object. """
        return self.to_pretty_frmt()

    def __setitem__(self, key, value):
        """
        Sets a metadata attribute.

        Parameters
        ----------
        key : str
            Metadata attribute.
        value : any
            Metadata value.

        """
        self._set_metadata(key, value)

    def __getitem__(self, item):
        """
        Returns metadata value according to the given metadata item.

        Parameters
        ----------
        item : str
            Metadata attribute.

        Returns
        -------
        any
            Metadata value.

        """
        return self._get_metadata(item)


def read_config(filepath):
    """
    Parse a metadata config file.

    Parameters
    ----------
    filepath : str
        Path to the metadata config file.

    Returns
    -------
    ds : dict
        Parsed metadata config file as a dictionary.

    """
    config = ConfigParser()
    config.optionxform = str
    config.read(filepath)

    ds = {}
    for section in config.sections():
        ds[section] = {}
        for item, value in config.items(section):
            if value.startswith('list'):
                value = value.replace(', ', ',')
                value = value.split(',')
                value.pop(0)
            ds[section][item] = value

    return ds
